package pqueue

import (
	"474420502.top/eson/structure/compare"
	"github.com/davecgh/go-spew/spew"
)

type tNode struct {
	children [2]*tNode
	parent   *tNode
	size     int
	value    interface{}
}

func (n *tNode) String() string {
	if n == nil {
		return "nil"
	}

	p := "nil"
	if n.parent != nil {
		p = spew.Sprint(n.parent.value)
	}
	return spew.Sprint(n.value) + "(" + p + "|" + spew.Sprint(n.size) + ")"
}

type vbTree struct {
	root    *tNode
	Compare compare.Compare
}

func newVBT(Compare compare.Compare) *vbTree {
	return &vbTree{Compare: Compare}
}

func (tree *vbTree) String() string {
	str := "AVLTree\n"
	if tree.root == nil {
		return str + "nil"
	}
	output(tree.root, "", true, &str)
	return str
}

func (tree *vbTree) Iterator() *vbtIterator {
	return initIterator(tree)
}

func (tree *vbTree) Size() int {
	if tree.root == nil {
		return 0
	}
	return tree.root.size
}

func (tree *vbTree) indexNode(idx int) *tNode {
	cur := tree.root
	if idx >= 0 {
		for cur != nil {
			ls := getSize(cur.children[0])
			if idx == ls {
				return cur
			} else if idx < ls {
				cur = cur.children[0]
			} else {
				idx = idx - ls - 1
				cur = cur.children[1]
			}
		}
	} else {
		idx = -idx - 1
		for cur != nil {
			rs := getSize(cur.children[1])
			if idx == rs {
				return cur
			} else if idx < rs {
				cur = cur.children[1]
			} else {
				idx = idx - rs - 1
				cur = cur.children[0]
			}
		}
	}
	return nil
}

func (tree *vbTree) Index(idx int) (interface{}, bool) {
	n := tree.indexNode(idx)
	if n != nil {
		return n.value, true
	}
	return nil, false
}

func (tree *vbTree) IndexRange(idx1, idx2 int) (result []interface{}, ok bool) { // 0 -1

	if idx1^idx2 < 0 {
		if idx1 < 0 {
			idx1 = tree.root.size + idx1 - 1
		} else {
			idx2 = tree.root.size + idx2 - 1
		}
	}

	if idx1 > idx2 {
		ok = true
		if idx1 >= tree.root.size {
			idx1 = tree.root.size - 1
			ok = false
		}

		n := tree.indexNode(idx1)
		iter := NewIterator(n)
		result = make([]interface{}, 0, idx1-idx2)
		for i := idx2; i <= idx1; i++ {
			if iter.Prev() {
				result = append(result, iter.Value())
			} else {
				ok = false
				return
			}
		}

		return

	} else {
		ok = true
		if idx2 >= tree.root.size {
			idx2 = tree.root.size - 1
			ok = false
		}

		if n := tree.indexNode(idx1); n != nil {
			iter := NewIterator(n)
			result = make([]interface{}, 0, idx2-idx1)
			for i := idx1; i <= idx2; i++ {
				if iter.Next() {
					result = append(result, iter.Value())
				} else {
					ok = false
					return
				}
			}

			return
		}

	}

	return nil, false
}

func (tree *vbTree) RemoveIndex(idx int) bool {
	n := tree.indexNode(idx)
	if n != nil {
		tree.removeNode(n)
		return true
	}
	return false
}

func (tree *vbTree) removeNode(n *tNode) {
	if tree.root.size == 1 {
		tree.root = nil
		// return n
		return
	}

	ls, rs := getChildrenSize(n)
	if ls == 0 && rs == 0 {
		p := n.parent
		p.children[getRelationship(n)] = nil
		// p.size -= n.size
		tree.fixSizeWithRemove(p)
		// return n
		return
	}

	var cur *tNode
	if ls > rs {
		cur = n.children[0]
		for cur.children[1] != nil {
			cur = cur.children[1]
		}

		cleft := cur.children[0]
		cur.parent.children[getRelationship(cur)] = cleft
		if cleft != nil {
			cleft.parent = cur.parent
		}

	} else {
		cur = n.children[1]
		for cur.children[0] != nil {
			cur = cur.children[0]
		}

		cright := cur.children[1]
		cur.parent.children[getRelationship(cur)] = cright
		if cright != nil {
			cright.parent = cur.parent
		}
	}

	cparent := cur.parent
	if cparent == n {
		cparent = cur
	}
	tree.replaceNodeMayRoot(n, cur)
	// 修改为interface 交换
	// n.value, cur.value = cur.value, n.value
	tree.fixSizeWithRemove(cparent)
	return
}

func (tree *vbTree) Remove(key interface{}) bool {

	if n, ok := tree.GetNode(key); ok {
		tree.removeNode(n)
		return true
	}
	// return nil
	return false
}

// Values 返回先序遍历的值
func (tree *vbTree) Values() []interface{} {
	mszie := 0
	if tree.root != nil {
		mszie = tree.root.size
	}
	result := make([]interface{}, 0, mszie)
	tree.Traversal(func(v interface{}) bool {
		result = append(result, v)
		return true
	}, LDR)
	return result
}

func (tree *vbTree) GetRange(k1, k2 interface{}) (result []interface{}) {
	c := tree.Compare(k2, k1)
	switch c {
	case 1:

		var min, max *tNode
		resultmin := tree.getArountNode(k1)
		resultmax := tree.getArountNode(k2)
		for i := 1; i < 3 && min == nil; i++ {
			min = resultmin[i]
		}

		for i := 1; i > -1 && max == nil; i-- {
			max = resultmax[i]
		}

		if max == nil {
			return []interface{}{}
		}

		result = make([]interface{}, 0, 16)

		iter := NewIterator(min)
		for iter.Next() {
			result = append(result, iter.Value())
			if iter.cur == max {
				break
			}
		}
	case -1:

		var min, max *tNode
		resultmin := tree.getArountNode(k2)
		resultmax := tree.getArountNode(k1)
		for i := 1; i < 3 && min == nil; i++ {
			min = resultmin[i]
		}
		for i := 1; i > -1 && max == nil; i-- {
			max = resultmax[i]
		}

		if min == nil {
			return []interface{}{}
		}

		result = make([]interface{}, 0, 16)

		iter := NewIterator(max)
		for iter.Prev() {
			result = append(result, iter.Value())
			if iter.cur == min {
				break
			}
		}
	case 0:
		if n, ok := tree.GetNode(k1); ok {
			return []interface{}{n.value}
		}
		return []interface{}{}
	}

	return
}

func (tree *vbTree) Get(key interface{}) (interface{}, bool) {
	n, ok := tree.GetNode(key)
	if ok {
		return n.value, true
	}
	return n, false
}

func (tree *vbTree) GetAround(key interface{}) (result [3]interface{}) {
	an := tree.getArountNode(key)
	for i, n := range an {
		if n != nil {
			result[i] = n.value
		}
	}
	return
}

func (tree *vbTree) getArountNode(key interface{}) (result [3]*tNode) {
	var last *tNode
	var lastc int

	for n := tree.root; n != nil; {
		last = n
		c := tree.Compare(key, n.value)
		switch c {
		case -1:
			n = n.children[0]
			lastc = c
		case 1:
			n = n.children[1]
			lastc = c
		case 0:
			iter := NewIterator(n)
			iter.Prev()
			for iter.Prev() {
				if tree.Compare(iter.cur.value, n.value) == 0 {
					n = iter.cur
				} else {
					break
				}
			}
			result[1] = n
			n = nil
		default:
			panic("Get Compare only is allowed in -1, 0, 1")
		}
	}

	switch lastc {
	case 1:

		if result[1] != nil {

			result[0] = getPrev(result[1], 1)
			result[2] = getNext(result[1], 1)
		} else {
			result[0] = last
			result[2] = getNext(last, 1)
		}

	case -1:

		if result[1] != nil {
			result[0] = getPrev(result[1], 1)
			result[2] = getNext(result[1], 1)
		} else {
			result[2] = last
			result[0] = getPrev(last, 1)
		}

	case 0:

		if result[1] == nil {
			return
		}
		result[0] = getPrev(result[1], 1)
		result[2] = getNext(result[1], 1)
	}
	return
}

func (tree *vbTree) GetNode(value interface{}) (*tNode, bool) {

	for n := tree.root; n != nil; {
		switch c := tree.Compare(value, n.value); c {
		case -1:
			n = n.children[0]
		case 1:
			n = n.children[1]
		case 0:
			iter := NewIterator(n)
			iter.Prev()
			for iter.Prev() {
				if tree.Compare(iter.cur.value, n.value) == 0 {
					n = iter.cur
				} else {
					break
				}
			}
			return n, true
		default:
			panic("Get Compare only is allowed in -1, 0, 1")
		}
	}
	return nil, false
}

func (tree *vbTree) Put(key interface{}) *tNode {

	node := &tNode{value: key, size: 1}
	if tree.root == nil {
		tree.root = node
		return node
	}

	cur := tree.root
	parent := cur.parent
	child := -1

	for {

		if cur == nil {
			parent.children[child] = node
			node.parent = parent

			fixed := parent.parent
			fsize := getSize(fixed)
			if fsize == 3 {
				lefts, rigths := getChildrenSize(fixed)
				tree.fix3Size(fixed, lefts, rigths)
			}
			return node
		}

		if cur.size > 8 {
			ls, rs := cur.children[0].size, cur.children[1].size
			factor := cur.size / 10 // or factor = 1
			if rs >= ls*2+factor || ls >= rs*2+factor {
				cur = tree.fixSize(cur, ls, rs)
			}
		}

		cur.size++
		parent = cur
		c := tree.Compare(key, cur.value)
		child = (c + 2) / 2
		cur = cur.children[child]
	}
}

type TraversalMethod int

const (
	// L = left R = right D = Value(dest)
	_ TraversalMethod = iota
	//DLR 先值 然后左递归 右递归 下面同理
	DLR
	//LDR 先从左边有序访问到右边 从小到大
	LDR
	// LRD 同理
	LRD

	// DRL 同理
	DRL

	// RDL 先从右边有序访问到左边 从大到小
	RDL

	// RLD 同理
	RLD
)

// Traversal 遍历的方法 默认是LDR 从小到大 Compare 为 l < r
func (tree *vbTree) Traversal(every func(v interface{}) bool, traversalMethod ...interface{}) {
	if tree.root == nil {
		return
	}

	method := LDR
	if len(traversalMethod) != 0 {
		method = traversalMethod[0].(TraversalMethod)
	}

	switch method {
	case DLR:
		var traverasl func(cur *tNode) bool
		traverasl = func(cur *tNode) bool {
			if cur == nil {
				return true
			}
			if !every(cur.value) {
				return false
			}
			if !traverasl(cur.children[0]) {
				return false
			}
			if !traverasl(cur.children[1]) {
				return false
			}
			return true
		}
		traverasl(tree.root)
	case LDR:
		var traverasl func(cur *tNode) bool
		traverasl = func(cur *tNode) bool {
			if cur == nil {
				return true
			}
			if !traverasl(cur.children[0]) {
				return false
			}
			if !every(cur.value) {
				return false
			}
			if !traverasl(cur.children[1]) {
				return false
			}
			return true
		}
		traverasl(tree.root)
	case LRD:
		var traverasl func(cur *tNode) bool
		traverasl = func(cur *tNode) bool {
			if cur == nil {
				return true
			}
			if !traverasl(cur.children[0]) {
				return false
			}
			if !traverasl(cur.children[1]) {
				return false
			}
			if !every(cur.value) {
				return false
			}
			return true
		}
		traverasl(tree.root)
	case DRL:
		var traverasl func(cur *tNode) bool
		traverasl = func(cur *tNode) bool {
			if cur == nil {
				return true
			}
			if !every(cur.value) {
				return false
			}
			if !traverasl(cur.children[0]) {
				return false
			}
			if !traverasl(cur.children[1]) {
				return false
			}
			return true
		}
		traverasl(tree.root)
	case RDL:
		var traverasl func(cur *tNode) bool
		traverasl = func(cur *tNode) bool {
			if cur == nil {
				return true
			}
			if !traverasl(cur.children[1]) {
				return false
			}
			if !every(cur.value) {
				return false
			}
			if !traverasl(cur.children[0]) {
				return false
			}
			return true
		}
		traverasl(tree.root)
	case RLD:
		var traverasl func(cur *tNode) bool
		traverasl = func(cur *tNode) bool {
			if cur == nil {
				return true
			}
			if !traverasl(cur.children[1]) {
				return false
			}
			if !traverasl(cur.children[0]) {
				return false
			}
			if !every(cur.value) {
				return false
			}
			return true
		}
		traverasl(tree.root)
	}
}

func (tree *vbTree) lrrotate3(cur *tNode) {
	const l = 1
	const r = 0
	// 1 right 0 left
	ln := cur.children[l]
	lrn := ln.children[r]

	if cur.parent != nil {
		if cur.parent.children[l] == cur {
			cur.parent.children[l] = lrn
		} else {
			cur.parent.children[r] = lrn
		}
	} else {
		tree.root = lrn
	}
	lrn.parent = cur.parent

	lrn.children[l] = ln
	lrn.children[l].parent = lrn

	lrn.children[r] = cur
	lrn.children[r].parent = lrn

	cur.children[l] = nil
	ln.children[r] = nil

	lrn.size = 3
	ln.size = 1
	cur.size = 1
}

func (tree *vbTree) lrrotate(cur *tNode) *tNode {

	const l = 1
	const r = 0

	// 1 right 0 left
	ln := cur.children[l]
	lrn := ln.children[r] // 待转换的节点
	lrln := lrn.children[l]
	lrrn := lrn.children[r]

	ln.children[r] = lrln
	if lrln != nil {
		ln.children[r].parent = ln
	}

	if cur.parent != nil {
		if cur.parent.children[l] == cur {
			cur.parent.children[l] = lrn
		} else {
			cur.parent.children[r] = lrn
		}
	} else {
		tree.root = lrn
	}
	lrn.parent = cur.parent

	lrn.children[l] = cur.children[l]
	lrn.children[l].parent = lrn

	lrn.children[r] = cur
	lrn.children[r].parent = lrn

	cur.children[l] = lrrn
	if lrrn != nil {
		cur.children[l].parent = cur
	}

	ln.size = getChildrenSumSize(ln) + 1
	cur.size = getChildrenSumSize(cur) + 1
	lrn.size = getChildrenSumSize(lrn) + 1

	return lrn
}

func (tree *vbTree) rlrotate3(cur *tNode) {
	const l = 0
	const r = 1
	// 1 right 0 left
	ln := cur.children[l]
	lrn := ln.children[r]

	if cur.parent != nil {
		if cur.parent.children[l] == cur {
			cur.parent.children[l] = lrn
		} else {
			cur.parent.children[r] = lrn
		}
	} else {
		tree.root = lrn
	}
	lrn.parent = cur.parent

	lrn.children[l] = ln
	lrn.children[l].parent = lrn

	lrn.children[r] = cur
	lrn.children[r].parent = lrn

	cur.children[l] = nil
	ln.children[r] = nil

	lrn.size = 3
	ln.size = 1
	cur.size = 1
}

func (tree *vbTree) rlrotate(cur *tNode) *tNode {

	const l = 0
	const r = 1

	// 1 right 0 left
	ln := cur.children[l]
	lrn := ln.children[r] // 待转换的节点
	lrln := lrn.children[l]
	lrrn := lrn.children[r]

	ln.children[r] = lrln
	if lrln != nil {
		ln.children[r].parent = ln
	}

	if cur.parent != nil {
		if cur.parent.children[l] == cur {
			cur.parent.children[l] = lrn
		} else {
			cur.parent.children[r] = lrn
		}
	} else {
		tree.root = lrn
	}
	lrn.parent = cur.parent

	lrn.children[l] = cur.children[l]
	lrn.children[l].parent = lrn

	lrn.children[r] = cur
	lrn.children[r].parent = lrn

	cur.children[l] = lrrn
	if lrrn != nil {
		cur.children[l].parent = cur
	}

	ln.size = getChildrenSumSize(ln) + 1
	cur.size = getChildrenSumSize(cur) + 1
	lrn.size = getChildrenSumSize(lrn) + 1

	return lrn
}

func (tree *vbTree) rrotate3(cur *tNode) {
	const l = 0
	const r = 1
	// 1 right 0 left
	ln := cur.children[l]

	if cur.parent != nil {
		if cur.parent.children[l] == cur {
			cur.parent.children[l] = ln
		} else {
			cur.parent.children[r] = ln
		}
	} else {
		tree.root = ln
	}
	ln.parent = cur.parent

	ln.children[r] = cur
	cur.parent = ln

	cur.children[l] = nil

	ln.size = 3
	cur.size = 1
}

func (tree *vbTree) rrotate(cur *tNode) *tNode {

	const l = 0
	const r = 1
	// 1 right 0 left
	ln := cur.children[l]
	lrn := ln.children[r] // 待移到另一则的节点

	if cur.parent != nil {
		if cur.parent.children[l] == cur {
			cur.parent.children[l] = ln
		} else {
			cur.parent.children[r] = ln
		}
	} else {
		tree.root = ln
	}
	ln.parent = cur.parent

	ln.children[r] = cur
	ln.children[r].parent = ln

	cur.children[l] = lrn
	if lrn != nil {
		cur.children[l].parent = cur
	}

	cur.size = getChildrenSumSize(cur) + 1
	ln.size = getChildrenSumSize(ln) + 1

	return ln
}

func (tree *vbTree) lrotate3(cur *tNode) {
	const l = 1
	const r = 0
	// 1 right 0 left
	ln := cur.children[l]

	if cur.parent != nil {
		if cur.parent.children[l] == cur {
			cur.parent.children[l] = ln
		} else {
			cur.parent.children[r] = ln
		}
	} else {
		tree.root = ln
	}
	ln.parent = cur.parent

	ln.children[r] = cur
	cur.parent = ln

	cur.children[l] = nil

	ln.size = 3
	cur.size = 1
}

func (tree *vbTree) lrotate(cur *tNode) *tNode {

	const l = 1
	const r = 0

	ln := cur.children[l]
	lrn := ln.children[r] // 待移到另一则的节点

	if cur.parent != nil {
		if cur.parent.children[l] == cur {
			cur.parent.children[l] = ln
		} else {
			cur.parent.children[r] = ln
		}
	} else {
		tree.root = ln
	}
	ln.parent = cur.parent

	ln.children[r] = cur
	ln.children[r].parent = ln

	cur.children[l] = lrn
	if lrn != nil {
		cur.children[l].parent = cur
	}

	cur.size = getChildrenSumSize(cur) + 1
	ln.size = getChildrenSumSize(ln) + 1

	return ln
}

func getChildrenSumSize(cur *tNode) int {
	return getSize(cur.children[0]) + getSize(cur.children[1])
}

func getChildrenSize(cur *tNode) (int, int) {
	return getSize(cur.children[0]), getSize(cur.children[1])
}

func getSize(cur *tNode) int {
	if cur == nil {
		return 0
	}
	return cur.size
}

func (tree *vbTree) fixSizeWithRemove(cur *tNode) {
	for cur != nil {
		cur.size--
		if cur.size > 8 {
			ls, rs := getChildrenSize(cur)
			factor := cur.size / 10 // or factor = 1
			if rs >= ls*2+factor || ls >= rs*2+factor {
				cur = tree.fixSize(cur, ls, rs)
			}
		}
		cur = cur.parent
	}
}

func (tree *vbTree) fix3Size(cur *tNode, lefts, rigths int) {
	if lefts > rigths {
		l := cur.children[0]
		llsize, lrsize := getChildrenSize(l)
		if lrsize > llsize {
			tree.rlrotate3(cur)
		} else {
			tree.rrotate3(cur)
		}
	} else {
		r := cur.children[1]
		rlsize, rrsize := getChildrenSize(r)
		if rlsize > rrsize {
			tree.lrrotate3(cur)
		} else {
			tree.lrotate3(cur)
		}
	}
}

func (tree *vbTree) fixSize(cur *tNode, lefts, rigths int) *tNode {
	if lefts > rigths {
		l := cur.children[0]
		llsize, lrsize := getChildrenSize(l)
		if lrsize > llsize {
			return tree.rlrotate(cur)
		} else {
			return tree.rrotate(cur)
		}
	} else {
		r := cur.children[1]
		rlsize, rrsize := getChildrenSize(r)
		if rlsize > rrsize {
			return tree.lrrotate(cur)
		} else {
			return tree.lrotate(cur)
		}
	}
}

func insertNode(parent *tNode, childidx int, newn *tNode) {
	child := parent.children[childidx]
	parent.children[childidx] = newn
	newn.parent = parent
	// 如果是旋转 还需要把 newn.children[childidx] 转到children的紧邻
	newn.children[childidx] = child
}

func replaceNodeNotRoot(oldn, newn *tNode) {
	if newn.parent == oldn {
		if oldn.children[0] == newn {
			oldn.children[0] = nil
		} else {
			oldn.children[1] = nil
		}
	}

	newn.children[0] = oldn.children[0]
	newn.children[1] = oldn.children[1]
	newn.parent = oldn.parent

	if oldn.parent.children[0] == oldn {
		oldn.parent.children[0] = newn
	} else {
		oldn.parent.children[1] = newn
	}
}

func (tree *vbTree) replaceNodeMayRoot(oldn, newn *tNode) {

	newn.children[0] = oldn.children[0]
	newn.children[1] = oldn.children[1]
	newn.size = oldn.size
	newn.parent = oldn.parent
	if oldn.parent != nil {
		if oldn.parent.children[0] == oldn {
			oldn.parent.children[0] = newn
		} else {
			oldn.parent.children[1] = newn
		}
	} else {
		tree.root = newn
	}
}

func (tree *vbTree) tailReplaceNode(oldn, newn *tNode) {
	newn.children[0] = oldn.children[0]
	newn.children[1] = oldn.children[1]
	newn.parent = oldn.parent
	if oldn.parent != nil {
		if oldn.parent.children[0] == oldn {
			oldn.parent.children[0] = newn
		} else {
			oldn.parent.children[1] = newn
		}
	} else {
		tree.root = newn
	}
}

func output(node *tNode, prefix string, isTail bool, str *string) {

	if node.children[1] != nil {
		newPrefix := prefix
		if isTail {
			newPrefix += "│   "
		} else {
			newPrefix += "    "
		}
		output(node.children[1], newPrefix, false, str)
	}
	*str += prefix
	if isTail {
		*str += "└── "
	} else {
		*str += "┌── "
	}

	*str += spew.Sprint(node.value) + "\n"

	if node.children[0] != nil {
		newPrefix := prefix
		if isTail {
			newPrefix += "    "
		} else {
			newPrefix += "│   "
		}
		output(node.children[0], newPrefix, true, str)
	}

}

func outputfordebug(node *tNode, prefix string, isTail bool, str *string) {

	if node.children[1] != nil {
		newPrefix := prefix
		if isTail {
			newPrefix += "│   "
		} else {
			newPrefix += "    "
		}
		outputfordebug(node.children[1], newPrefix, false, str)
	}
	*str += prefix
	if isTail {
		*str += "└── "
	} else {
		*str += "┌── "
	}

	suffix := "("
	parentv := ""
	if node.parent == nil {
		parentv = "nil"
	} else {
		parentv = spew.Sprint(node.parent.value)
	}
	suffix += parentv + "|" + spew.Sprint(node.size) + ")"
	*str += spew.Sprint(node.value) + suffix + "\n"

	if node.children[0] != nil {
		newPrefix := prefix
		if isTail {
			newPrefix += "    "
		} else {
			newPrefix += "│   "
		}
		outputfordebug(node.children[0], newPrefix, true, str)
	}
}

func (tree *vbTree) debugString() string {
	str := "AVLTree\n"
	if tree.root == nil {
		return str + "nil"
	}
	outputfordebug(tree.root, "", true, &str)
	return str
}
